/*
 * This file is a part of the CaosDB Project.
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/transaction.h"
#include "caosdb/entity/v1alpha1/main.grpc.pb.h" // for EntityTransactionS...
#include "caosdb/entity/v1alpha1/main.pb.h"      // for SingleRetrieveRequest
#include "caosdb/logging.h"
#include "caosdb/protobuf_helper.h"               // for get_arena
#include "caosdb/status_code.h"                   // for StatusCode, AUTHEN...
#include "google/protobuf/arena.h"                // for Arena
#include "grpcpp/grpcpp.h"                        // for CompletionQueue
#include "grpcpp/impl/codegen/async_unary_call.h" // for ClientAsyncRespons...
#include "grpcpp/impl/codegen/client_context.h"   // for ClientContext
#include "grpcpp/impl/codegen/completion_queue.h" // for CompletionQueue
#include "grpcpp/impl/codegen/status.h"           // for Status
#include "grpcpp/impl/codegen/status_code_enum.h" // for StatusCode, UNAUTH...
#include <cassert>                                // for assert
#include <map>                                    // for map
#include <memory>                                 // for allocator, unique_ptr
#include <stdexcept>                              // for out_of_range
#include <utility>                                // for move

namespace caosdb {

// TODO(tf) move to another source file.
auto get_status_description(int code) -> const std::string & {
  static const std::string MISSING_DESCRIPTION = "MISSING DESCRIPTION";
  static const std::map<int, std::string> descriptions = {
    {StatusCode::INITIAL,
     "The transaction has just been intialized. It has not been executed yet "
     "and clients can still change it and add sub-transactions."},
    {StatusCode::GO_ON,
     "The transaction has a transaction_type yet and clients may add matching "
     "sub-transaction or execute it right-away."},
    {StatusCode::READY,
     "The transaction is ready for execution and cannot be changed anymore."},
    {StatusCode::EXECUTING, "The transaction is currently being executed."},
    {StatusCode::SUCCESS, "The action was successful"},
    {StatusCode::CONNECTION_ERROR,
     "The attempt to execute this transaction was not successful because the "
     "connection to the server could not be established."},
    {StatusCode::AUTHENTICATION_ERROR,
     "The attempt to execute this transaction has not been executed at all "
     "because the authentication did not succeed."},
    {StatusCode::GENERIC_RPC_ERROR,
     "The attempt to execute this transaction was not successful because an "
     "error occured in the transport or RPC protocol layer."},
    {StatusCode::GENERIC_ERROR,
     "An error occured. Please review the logs for more information."},
    {StatusCode::GENERIC_TRANSACTION_ERROR,
     "The transaction terminated unsuccessfully with transaction errors."},
    {StatusCode::CONFIGURATION_ERROR,
     "An error occurred during the configuration of the ConfigurationManager."},
    {StatusCode::UNKNOWN_CONNECTION_ERROR,
     "The ConnectionManager does not know any connection of this name."},
    {StatusCode::TRANSACTION_STATUS_ERROR,
     "The Transaction is in a wrong state for the attempted action."},
    {StatusCode::TRANSACTION_TYPE_ERROR,
     "The Transaction has a transaction type which does not allow the "
     "attempted action."},
    {StatusCode::ORIGINAL_ENTITY_MISSING_ID,
     "The attempt to update this entity failed because this entity does have "
     "an id. This is the case when you did not retrieve it before applying any "
     "changes and instantiated the Entity class explicitely."},
    {StatusCode::UNSUPPORTED_FEATURE,
     "This feature is not available in the this client implementation."}};
  try {
    return descriptions.at(code);
  } catch (const std::out_of_range &exc) {
    return MISSING_DESCRIPTION;
  }
}

} // namespace caosdb

namespace caosdb::transaction {
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::entity::v1alpha1::MultiTransactionRequest;
using caosdb::entity::v1alpha1::MultiTransactionResponse;
using WrappedResponseCase =
  caosdb::entity::v1alpha1::TransactionResponse::WrappedResponseCase;
using caosdb::utility::get_arena;
using grpc::ClientAsyncResponseReader;
using ProtoEntity = caosdb::entity::v1alpha1::Entity;
using grpc::CompletionQueue;

MultiResultSet::MultiResultSet(MultiTransactionResponse *response) {
  auto *responses = response->mutable_responses();
  for (auto sub_response : *responses) {
    switch (sub_response.wrapped_response_case()) {
    case WrappedResponseCase::kRetrieveResponse:
      this->entities.push_back(std::make_unique<Entity>(
        sub_response.mutable_retrieve_response()->release_entity()));
      break;
    case WrappedResponseCase::kInsertResponse:
      this->entities.push_back(
        std::make_unique<Entity>(sub_response.release_insert_response()));
      break;
    case WrappedResponseCase::kDeleteResponse:
      this->entities.push_back(
        std::make_unique<Entity>(sub_response.release_insert_response()));
      break;
    default:
      // TODO(tf) Updates
      break;
    }
  }
}

[[nodiscard]] auto UniqueResult::GetEntity() const -> const Entity & {
  const Entity *result = this->entity.get();
  return *result;
}

Transaction::Transaction(
  std::shared_ptr<EntityTransactionService::Stub> service_stub)
  : request(google::protobuf::Arena::CreateMessage<MultiTransactionRequest>(
      get_arena())),
    response(google::protobuf::Arena::CreateMessage<MultiTransactionResponse>(
      get_arena())) {
  this->service_stub = std::move(service_stub);
}

auto Transaction::RetrieveById(const std::string &id) noexcept -> StatusCode {
  ASSERT_CAN_ADD_RETRIEVAL

  auto *sub_request = this->request->add_requests();
  sub_request->mutable_retrieve_request()->set_id(id);

  return StatusCode::INITIAL;
}

auto Transaction::Query(const std::string &query) noexcept -> StatusCode {
  ASSERT_CAN_ADD_RETRIEVAL

  auto *sub_request = this->request->add_requests();
  sub_request->mutable_retrieve_request()->mutable_query()->set_query(query);

  return StatusCode::INITIAL;
}

auto Transaction::DeleteById(const std::string &id) noexcept -> StatusCode {
  ASSERT_CAN_ADD_DELETION

  auto *sub_request = this->request->add_requests();
  sub_request->mutable_delete_request()->set_id(id);

  return StatusCode::INITIAL;
}

auto Transaction::InsertEntity(Entity *entity) noexcept -> StatusCode {
  ASSERT_CAN_ADD_INSERTION

  auto *sub_request = this->request->add_requests();
  auto *proto_entity = sub_request->mutable_insert_request();

  // copy the original entity for the transaction
  entity->CopyTo(proto_entity);
  return StatusCode::INITIAL;
}

auto Transaction::UpdateEntity(Entity *entity) noexcept -> StatusCode {
  ASSERT_CAN_ADD_UPDATE

  auto *sub_request = this->request->add_requests();
  auto *proto_entity = sub_request->mutable_update_request();

  entity->Switch(proto_entity);

  return StatusCode::INITIAL;
}

auto Transaction::Execute() -> TransactionStatus {
  ExecuteAsynchronously();
  auto status = WaitForIt();
  status.ThrowExceptionIfError();
  return status;
}

auto Transaction::ExecuteAsynchronously() noexcept -> StatusCode {
  if (!IsStatus(TransactionStatus::INITIAL())) {
    return StatusCode::TRANSACTION_STATUS_ERROR;
  }
  switch (this->transaction_type) {
  case MIXED_WRITE:
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(
      logger_name, StatusCode::UNSUPPORTED_FEATURE,
      "MIXED_WRITE UNSUPPORTED: The current implementation does not support "
      "mixed write transactions (containing insertions, deletions, and updates "
      "in one transaction).")
  case MIXED_READ_AND_WRITE:
    CAOSDB_LOG_ERROR_AND_RETURN_STATUS(
      logger_name, StatusCode::UNSUPPORTED_FEATURE,
      "MIXED_WRITE UNSUPPORTED: The current implementation does not support "
      "mixed read and write transactions (containing retrievals, insertions, "
      "deletions, and updates in one transaction).")
  default:
    break;
  }
  this->status = TransactionStatus::EXECUTING();

  grpc::Status grpc_status;
  CompletionQueue cq;

  grpc::ClientContext context;
  std::unique_ptr<ClientAsyncResponseReader<MultiTransactionResponse>> rpc(
    this->service_stub->PrepareAsyncMultiTransaction(&context, *(this->request),
                                                     &cq));
  rpc->StartCall();

  int tag = 1;
  void *send_tag = static_cast<void *>(&tag);
  rpc->Finish(this->response, &grpc_status, send_tag);
  void *recv_tag = nullptr;
  bool ok = false;

  // TODO(tf) make this actually asynchronous by moving this to WaitForIt()
  cq.Next(&recv_tag, &ok);
  assert(recv_tag == send_tag);
  assert(ok);

  if (!grpc_status.ok()) {
    switch (grpc_status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      this->status = TransactionStatus::AUTHENTICATION_ERROR();
      break;
    case grpc::StatusCode::UNAVAILABLE:
      this->status = TransactionStatus::CONNECTION_ERROR();
      break;
    default:
      auto error_details = std::to_string(grpc_status.error_code()) + " - " +
                           grpc_status.error_message();
      this->status = TransactionStatus::RPC_ERROR(error_details);
    }
  } else {
    this->status = TransactionStatus::SUCCESS();
  }
  return StatusCode::EXECUTING;
}

auto Transaction::WaitForIt() const noexcept -> TransactionStatus {
  if (this->response->responses_size() == 1) {
    auto *responses = this->response->mutable_responses(0);
    switch (responses->wrapped_response_case()) {
    case WrappedResponseCase::kRetrieveResponse: {
      auto *entity = responses->mutable_retrieve_response()->release_entity();
      if (!entity->errors().empty()) {
        this->status = TransactionStatus::TRANSACTION_ERROR(
          "The request returned with errors.");
      }
      this->result_set = std::make_unique<UniqueResult>(entity);
    } break;
    case WrappedResponseCase::kUpdateResponse: {
      auto *updatedIdResponse = responses->mutable_update_response();
      if (!updatedIdResponse->entity_errors().empty()) {
        this->status = TransactionStatus::TRANSACTION_ERROR(
          "The request returned with errors.");
      }
      this->result_set = std::make_unique<UniqueResult>(updatedIdResponse);
    } break;
    case WrappedResponseCase::kInsertResponse: {
      auto *insertedIdResponse = responses->mutable_insert_response();
      if (!insertedIdResponse->entity_errors().empty()) {
        this->status = TransactionStatus::TRANSACTION_ERROR(
          "The request returned with errors.");
      }
      this->result_set = std::make_unique<UniqueResult>(insertedIdResponse);
    } break;
    case WrappedResponseCase::kDeleteResponse: {
      auto *deletedIdResponse = responses->mutable_delete_response();
      if (!deletedIdResponse->entity_errors().empty()) {
        this->status = TransactionStatus::TRANSACTION_ERROR(
          "The request returned with errors.");
      }
      this->result_set = std::make_unique<UniqueResult>(deletedIdResponse);
    } break;
    default:
      // TODO(tf) Error and Update
      break;
    }
  } else {
    this->result_set = std::make_unique<MultiResultSet>(this->response);
  }

  return this->status;
}

} // namespace caosdb::transaction
